@@ -3,6 +3,7 @@
 #![feature(maybe_uninit_write_slice)]
 #![feature(maybe_uninit_slice)]
 #![feature(step_trait)]
+#![feature(new_uninit)]
 #![allow(dead_code)]
 
 //! A library that can be used as a building block for high-performant graph
@@ -46,7 +47,7 @@
 //! ```
 //! use graph::prelude::*;
 //!
-//! let graph: DirectedCSRGraph<usize> = GraphBuilder::new()
+//! let graph: DirectedCsrGraph<usize> = GraphBuilder::new()
 //!     .edges(vec![(0, 1), (0, 2), (1, 2), (1, 3), (2, 3)])
 //!     .build();
 //!
@@ -66,7 +67,7 @@
 //! ```
 //! use graph::prelude::*;
 //!
-//! let graph: UndirectedCSRGraph<u32> = GraphBuilder::new()
+//! let graph: UndirectedCsrGraph<u32> = GraphBuilder::new()
 //!     .edges(vec![(0, 1), (0, 2), (1, 2), (1, 3), (2, 3)])
 //!     .build();
 //!
@@ -91,8 +92,8 @@
 //!     .iter()
 //!     .collect::<PathBuf>();
 //!
-//! let graph: DirectedCSRGraph<usize> = GraphBuilder::new()
-//!     .csr_option(CSROption::Sorted)
+//! let graph: DirectedCsrGraph<usize> = GraphBuilder::new()
+//!     .csr_layout(CsrLayout::Sorted)
 //!     .file_format(EdgeListInput::default())
 //!     .path(path)
 //!     .build()
@@ -121,6 +122,8 @@ pub mod index;
 pub mod input;
 pub mod prelude;
 
+use std::convert::Infallible;
+
 use crate::index::Idx;
 
 use thiserror::Error;
@@ -128,14 +131,27 @@ use thiserror::Error;
 #[derive(Error, Debug)]
 pub enum Error {
     #[error("error while loading graph")]
-    LoadGraph {
+    IoError {
         #[from]
         source: std::io::Error,
     },
+    #[error("error while parsing GDL input")]
+    GdlError {
+        #[from]
+        source: gdl::graph::GraphHandlerError,
+    },
     #[error("invalid partitioning")]
     InvalidPartitioning,
     #[error("number of node values must be the same as node count")]
     InvalidNodeValues,
+    #[error("invalid id size, expected {expected:?} bytes, got {actual:?} bytes")]
+    InvalidIdType { expected: String, actual: String },
+}
+
+impl From<Infallible> for Error {
+    fn from(_: Infallible) -> Self {
+        unreachable!()
+    }
 }
 
 /// A graph is a tuple `(N, E)`, where `N` is a set of nodes and `E` a set of
@@ -212,8 +228,8 @@ mod tests {
 
     use crate::{
         builder::GraphBuilder,
-        graph::csr::{CSROption, DirectedCSRGraph, UndirectedCSRGraph},
-        input::edgelist::EdgeListInput,
+        graph::csr::{CsrLayout, DirectedCsrGraph, UndirectedCsrGraph},
+        input::{binary::BinaryInput, edgelist::EdgeListInput},
     };
 
     use super::*;
@@ -221,21 +237,26 @@ mod tests {
     #[test]
     fn should_compile_test() {
         fn inner_test() -> Result<(), Error> {
-            let _g: DirectedCSRGraph<usize> = GraphBuilder::new()
+            let _g: DirectedCsrGraph<usize> = GraphBuilder::new()
                 .file_format(EdgeListInput::default())
                 .path("graph")
                 .build()?;
 
-            let _g: DirectedCSRGraph<_> = GraphBuilder::new()
+            let _g: DirectedCsrGraph<_> = GraphBuilder::new()
                 .file_format(EdgeListInput::<usize>::default())
                 .path("graph")
                 .build()?;
 
-            let _g: UndirectedCSRGraph<usize> = GraphBuilder::new()
+            let _g: UndirectedCsrGraph<usize> = GraphBuilder::new()
                 .file_format(EdgeListInput::default())
                 .path("graph")
                 .build()?;
 
+            let _g: DirectedCsrGraph<usize> = GraphBuilder::new()
+                .file_format(BinaryInput::<usize>::default())
+                .path("graph")
+                .build()?;
+
             Ok(())
         }
 
@@ -269,7 +290,7 @@ mod tests {
             .collect::<PathBuf>();
 
         let graph = GraphBuilder::new()
-            .csr_option(CSROption::Sorted)
+            .csr_layout(CsrLayout::Sorted)
             .file_format(EdgeListInput::default())
             .path(path)
             .build()
@@ -285,7 +306,7 @@ mod tests {
             .collect::<PathBuf>();
 
         let graph = GraphBuilder::new()
-            .csr_option(CSROption::Sorted)
+            .csr_layout(CsrLayout::Sorted)
             .file_format(EdgeListInput::default())
             .path(path)
             .build()
@@ -301,7 +322,7 @@ mod tests {
             .collect::<PathBuf>();
 
         let graph = GraphBuilder::new()
-            .csr_option(CSROption::Sorted)
+            .csr_layout(CsrLayout::Sorted)
             .file_format(EdgeListInput::default())
             .path(path)
             .build()
@@ -317,7 +338,7 @@ mod tests {
             .collect::<PathBuf>();
 
         let graph = GraphBuilder::new()
-            .csr_option(CSROption::Sorted)
+            .csr_layout(CsrLayout::Sorted)
             .file_format(EdgeListInput::default())
             .path(path)
             .build()
@@ -326,7 +347,7 @@ mod tests {
         assert_undirected_graph::<u32>(graph);
     }
 
-    fn assert_directed_graph<Node: Idx>(g: DirectedCSRGraph<Node>) {
+    fn assert_directed_graph<Node: Idx>(g: DirectedCsrGraph<Node>) {
         assert_eq!(g.node_count(), Node::new(3));
         assert_eq!(g.edge_count(), Node::new(2));
 
@@ -347,7 +368,7 @@ mod tests {
         assert_eq!(g.in_neighbors(Node::new(2)), &[Node::new(0)]);
     }
 
-    fn assert_undirected_graph<Node: Idx>(g: UndirectedCSRGraph<Node>) {
+    fn assert_undirected_graph<Node: Idx>(g: UndirectedCsrGraph<Node>) {
         assert_eq!(g.node_count(), Node::new(3));
         assert_eq!(g.edge_count(), Node::new(2));
 
