@@ -1,11 +1,12 @@
 pub use crate::builder::GraphBuilder;
 
-pub use crate::graph::csr::CSROption;
-pub use crate::graph::csr::DirectedCSRGraph;
-pub use crate::graph::csr::UndirectedCSRGraph;
+pub use crate::graph::csr::CsrLayout;
+pub use crate::graph::csr::DirectedCsrGraph;
+pub use crate::graph::csr::UndirectedCsrGraph;
 
 pub use crate::graph_ops::DegreePartitionOp;
-pub use crate::graph_ops::ForEachNodeOp;
+pub use crate::graph_ops::ForEachNodeParallelByPartitionOp;
+pub use crate::graph_ops::ForEachNodeParallelOp;
 pub use crate::graph_ops::InDegreePartitionOp;
 pub use crate::graph_ops::OutDegreePartitionOp;
 pub use crate::graph_ops::RelabelByDegreeOp;
